package com.webchat.aigc.llm;

import com.fasterxml.jackson.core.type.TypeReference;
import com.webchat.common.bean.APIResponseBean;
import com.webchat.common.bean.APIResponseBeanUtil;
import com.webchat.common.constants.MilvusConstants;
import com.webchat.common.enums.PromptTemplateEnum;
import com.webchat.common.service.FreeMarkEngineService;
import com.webchat.common.util.JsonUtil;
import com.webchat.domain.vo.llm.ChatCompletionMessage;
import com.webchat.domain.vo.llm.ChatMessageRole;
import com.webchat.domain.vo.request.mess.ChatMessageRequestVO;
import com.webchat.domain.vo.request.search.MilvusSearchRequestVO;
import com.webchat.domain.vo.response.search.ArticleMilvusSearchResponse;
import com.webchat.rmi.search.VectorSearchEngineClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cloud.context.config.annotation.RefreshScope;
import org.springframework.stereotype.Service;
import org.springframework.web.servlet.mvc.method.annotation.SseEmitter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


@RefreshScope
@Service
public class AiBotQAService {


    @Autowired
    private VectorSearchEngineClient vectorSearchEngineClient;

    @Autowired
    private FreeMarkEngineService freeMarkEngineService;

    @Value("${llm.config.model}")
    private String model;

    /**
     * 向量相似度搜索默认取相似度最高前两条数据
     */
    private static final int SEARCH_TOP_K = 2;

    /**
     * 相似度限制条件
     */
    private static final float MIN_SCORE = 0.8f;

    private static final String AIBOT_ROLE = "你是Chat4j项目的Ai助手，任务是根据用户给出的参考数据，基于输入query，精准的给出答案。";

    /**
     * 我的ai助手问答场景实现 （RAG）
     *
     * @param sseEmitter
     * @param chatMessage
     * @return
     */
    public String chat(SseEmitter sseEmitter, ChatMessageRequestVO chatMessage) {

        String query = chatMessage.getMessage();
        String senderId = chatMessage.getSenderId();
        /**
         * 基于公众号文章相似度实现搜索
         */
        List<ArticleMilvusSearchResponse> searchResponses = this.search(senderId, query);

        /**
         * prompt工程构件需要大模型处理指令
         */
        Map<String, Object> vars = new HashMap<>();
        vars.put("searchResponses", searchResponses);
        vars.put("input", query);
        String prompt;
        try {
            prompt = freeMarkEngineService.getContentByTemplate(PromptTemplateEnum.AIBOT_RAG.getPath(), vars);
        } catch (Exception e) {
            this.sseSend(sseEmitter, "prompt构建失败");
            e.printStackTrace();
            return null;
        }
        /**
         * 增强
         */
        // 使用工厂模式，取对应的LLM服务，来完成对话
        AbstractLLMChatService abstractLLMChatService = LLMServiceFactory.getLLMService(model);
        final List<ChatCompletionMessage> messageList = Arrays.asList(
                new ChatCompletionMessage(ChatMessageRole.SYSTEM.value(), AIBOT_ROLE),
                new ChatCompletionMessage(ChatMessageRole.USER.value(), prompt));
        String aiResponse;
        try {
            aiResponse = abstractLLMChatService.chat(sseEmitter, messageList);
        } catch (Exception e) {
            this.sseSend(sseEmitter, "服务器繁忙");
            return null;
        }
        return aiResponse;
    }

    /**
     * RPC 远程请求我们的搜索引擎服务，基于向量相似度检索，搜索topK条公众号文章片段
     *
     * @param userId
     * @param query
     * @return
     */
    private List<ArticleMilvusSearchResponse> search(String userId, String query) {
        String biz = MilvusConstants.SearchBiz.SEARCH_ARTICLE.getBiz();
        MilvusSearchRequestVO req = MilvusSearchRequestVO.builder()
                .userId(userId)
                .query(query)
                .score(MIN_SCORE)
                .topK(SEARCH_TOP_K)
                .build();
        APIResponseBean<Object> response = vectorSearchEngineClient.search(biz, req);
        if (APIResponseBeanUtil.isOk(response)) {
            return JsonUtil.fromJson(JsonUtil.toJsonString(response.getData()), new TypeReference<List<ArticleMilvusSearchResponse>>() { });
        }
        return Collections.emptyList();
    }

    /**
     * sse 消息推送
     *
     * @param message
     */
    public static void sseSend(SseEmitter sseEmitter, String message) {
        try {
            sseEmitter.send(message);
            sseEmitter.send("finished");
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }
}
